/**
 * Program Description: This purpose of this program is to translate a <code>SBML</code> file
 *                      to <code>SAL</code> file.
 * @author:             Sri Paladugu.
 * @version:            0.0a.
 * Date Compiled:       07/15/2003.
 * Input:               SBML File.
 * Output:              SAL File
 */

// History
// Sep17: checks in equals(qName, "event")
// Sep23: in isFunctionDefinition: if (isMath || !isMath) EVERYWHERE
// 	because jigcell doesn't create <math:math> after functionDefinition
// Sep23: modelId fixed. either id, or name, or "Model". Removes space too.

// import sbmlClasses.*;
import java.io.*;
import java.util.*;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.XMLReader;
import org.xml.sax.SAXException;
import org.xml.sax.Attributes;
import org.xml.sax.helpers.DefaultHandler;

/**
 * A <code>Sbml_Sal_Translator</code> object represents a SBML to SAL Translator. It extends
 * DefaultHandler which is an empty implementation of ContentHandler interface. ContentHanler
 * is a SAX interface which will be used by the xml parser to communicate with the object
 * Sbml_Sal_Translator.
 */   
public class Sbml_Sal_Translator extends DefaultHandler
{
  private ArrayList listOfCompartments        = new ArrayList();
  private ArrayList listOfSpecies             = new ArrayList();
  private ArrayList listOfReactions           = new ArrayList();
  private ArrayList listOfReactants           = new ArrayList();
  private ArrayList listOfProducts            = new ArrayList();
  private ArrayList listOfModifiers           = new ArrayList();
  private ArrayList listOfParameters          = new ArrayList();
  private ArrayList listOfGlobalParameters    = new ArrayList();
  private ArrayList listOfAlgebraicRules      = new ArrayList();
  private ArrayList listOfAssignmentRules     = new ArrayList();
  private ArrayList listOfRateRules           = new ArrayList();
  private ArrayList listOfFunctionDefinitions = new ArrayList();
  private ArrayList listOfFunctionVariables   = new ArrayList();
  private ArrayList listOfEvents              = new ArrayList();
  private ArrayList listOfEventAssignments    = new ArrayList();
  private ArrayList listOfEventVariables      = new ArrayList();
  private ArrayList listOfTriggerVariables    = new ArrayList();
  private ArrayList listOfRuleVariables       = new ArrayList();

  private boolean isListOfReactants           = false;
  private boolean isListOfProducts            = false;
  private boolean isListOfModifiers           = false;
  private boolean isKineticLaw                = false;
  private boolean isMath                      = false;
  private boolean isMathIdentifier            = false;
  private boolean isSpeciesReference          = false;
  private boolean isAssignmentRule            = false;
  private boolean isAlgebraicRule             = false;
  private boolean isRateRule                  = false;
  private boolean isFunctionDefinition        = false;
  private boolean isFunctionVariable          = false;
  private boolean isEventAssignment           = false;
  private boolean isTrigger                   = false;
  private boolean ignoreFunctionDefinitions   = true;		//ASHISH
  private boolean verbose		      = false;

  private static String modelId;
  private static String reactionName;
  private static String reactionReversible;
  private static String formula;
  private static String speciesValue;
  private static String stoichiometryValue;
  private static String ruleVariable;
  private static String functionId;
  private static String functionVariable;
  private static String eventAssignment;
  private static String eventVariable;
  private static String triggerExpression;

  private static StringBuffer mathIdentifier = new StringBuffer();

  public static void main(String[] args) throws Exception
  {
    int index = 0;

    System.out.println("SBML2SAL Translator...");
    if (args.length == 0 || args.length > 2) {
	System.out.println("Usage: sbml2sal [-v] <sbml-file.xml>");
    } else {
	index = args.length - 1;
    	Sbml_Sal_Translator readerObj = new Sbml_Sal_Translator();
	if (args.length > 1) {
	   if (args[0].equals("-v")) readerObj.verbose = true;
	   else System.out.println("Usage: sbml2sal [-v] <sbml-file.xml>");
	} 
    	readerObj.read(args[index]);
    }
    System.out.println("SBML2SAL Done.");
  }

  /** 
   * This method provides the SAX Parser with the document to be parsed, by invoking parse() method.
   * @param fileName     The file to be parsed
   */
  public void read (String fileName) throws Exception
  {
    XMLReader readerObj = XMLReaderFactory.createXMLReader("org.apache.xerces.parsers.SAXParser");               
    readerObj.setContentHandler (this);
    readerObj.parse (fileName);
  }
  
  /**
   * This method catches <em>start of document<em> event.
   * @throws SAXException
   */
  public void startDocument() throws SAXException
  {
    System.out.println("Started Reading ...");
  }
 
  /**
   * This method catches <em>end of document<em> event.
   * @throws SAXException
   */
  public void endDocument() throws SAXException
  {
    printCompartments();
    System.out.println("... Finished Reading");
    try
    {
       writeToSalFile();
    }
    catch (Exception e)
    {
      System.err.println("Unable to write to Sal File");
      e.printStackTrace();
    }
  }

  /**
   * This method creates a file with .sal extension and writes the data in sal Format
   * @throws Exception
   */
  public void writeToSalFile() throws Exception
  {
     PrintWriter out = new PrintWriter (new FileWriter(modelId + ".sal"), true);
     out.print(modelId);                                                         // Begin of writing CONTEXT
     out.println(" : CONTEXT =");          
     out.println("BEGIN\n");                                                
   
     ArrayList listOfGlobalAssignmentRules       = new ArrayList();
     ArrayList listOfGlobalAlgebraicRules        = new ArrayList();
     ArrayList listOfGlobalRateRules             = new ArrayList();
     ArrayList listOfConstantGlobalParameters    = new ArrayList();
     ArrayList listOfNonConstantGlobalParameters = new ArrayList();
     boolean   isGlobalRule                      = true;
     boolean   isGlobalParameter                 = false;
     String    ruleVariable                      = "";
     
     /* Begin Writing Function Definitions */
     if (!ignoreFunctionDefinitions) {
     for (int i=0; i<listOfFunctionDefinitions.size(); i++)
     {  
         FunctionDefinition f = (FunctionDefinition)  listOfFunctionDefinitions.get(i);
         out.print(f.getFunctionId() + "(" );
         
         for (int j=0; j<f.getListOfFunctionVariablesSize(); j++)
         {
             out.print(f.getListOfFunctionVariables().get(j) );

             if (j < f.getListOfFunctionVariablesSize()-1)
             {
                out.print(",");
             }
             else
             {
                out.print(": REAL)");
             }
         }
         out.println(" : REAL = ");

         out.println("\t" + f.getFunctionExpression() + ";");
     } /* End of Function Definitions section */
     }

     ArrayList listOfLocalSpecies    = new ArrayList();
     ArrayList listOfLocalReactions  = new ArrayList();
     boolean   reactionAdded         = false;

     for (int i=0; i<listOfCompartments.size(); i++)                            // Begin of writing MODULES
     {
         Compartment compartment = (Compartment)listOfCompartments.get(i);
         
         for (int j=0; j<listOfSpecies.size(); j++)                            // Get all the species corresponding to the
         {                                                                     // Compartment/Module.
       	     Species species = (Species)listOfSpecies.get(j);

             if ( species.getCompartment().equals( compartment.getId() ) )
             {
                listOfLocalSpecies.add(species); 
             }
         }
	 compartment.setLocalSpecies(listOfLocalSpecies, listOfRateRules);

         for (int j=0; j<listOfReactions.size(); j++)                            // Get all the reactions corresponding to the 
         {                                                                       // Compartment/Module to a local List Array.
             Reaction reaction              = (Reaction)listOfReactions.get(j);
                      reactionAdded         = false;
             
             /** If a reactant-species matches a species in Local Species list Array then the reaction belongs to local
                 Compartment/Module */
             for (int k=0; k<reaction.getListOfReactantsSize() && (!reactionAdded); k++)
             {
                 SpeciesReference speciesRef = (SpeciesReference) reaction.getListOfReactants().get(k);
               
                 for (int l=0; l<listOfLocalSpecies.size() && (!reactionAdded); l++)
                 {
                     Species localSpecies = (Species) listOfLocalSpecies.get(l);
                     if ( localSpecies.getId().equals(speciesRef.getSpecies()) )
                     {
                        listOfLocalReactions.add(reaction);
                        reactionAdded = true;
                     }
                 }                                                                          
           
             }
            
             /** If a product-species matches a species in Local Species list Array then the reaction belongs to local
                 Compartment/Module. Note: Search the Product List Array only if the reaction is not already added to 
                 Local reactions list Array */
             for (int k=0; k<reaction.getListOfProductsSize() && (!reactionAdded); k++)
             {
                 SpeciesReference speciesRef = (SpeciesReference) reaction.getListOfProducts().get(k);
              
                 for (int l=0; l<listOfLocalSpecies.size() && (!reactionAdded); l++)
                 {
                     Species localSpecies = (Species) listOfLocalSpecies.get(l);
                     if ( localSpecies.getId().equals(speciesRef.getSpecies()) )
                     {
                        listOfLocalReactions.add(reaction);
                        reactionAdded = true;
                     }
                 }                                                                          
              
             }
             /** If a modifier-species matches a species in Local Species list Array then the reaction belongs to local
                 Compartment/Module. Note: Search the Modifier List Array only if the reaction is not already added to 
                 Local reactions list Array */
               
             for (int k=0; k<reaction.getListOfModifiersSize() && (!reactionAdded); k++)
             {
                 ModifierSpeciesReference modifierRef = (ModifierSpeciesReference) reaction.getListOfModifiers().get(k);
              
                 for (int l=0; l<listOfLocalSpecies.size() && (!reactionAdded); l++)
                 {
                     Species localSpecies = (Species) listOfLocalSpecies.get(l);
                     if ( localSpecies.getId().equals(modifierRef.getSpecies()) )
                     {
                        listOfLocalReactions.add(reaction);
                        reactionAdded = true;
                     }
                 }                                                                          
              
             }

         
         }
	 compartment.setLocalReactions(listOfLocalReactions);
       
         ArrayList listOfLocalParameters = new ArrayList();
         for (int j=0; j<listOfLocalReactions.size(); j++)                    // Get all the local parameters from each and
         {                                                                    // every reaction in a Module/Compartment and
             Reaction   reaction   = (Reaction)listOfLocalReactions.get(j);   // store in a temporary list array.
             KineticLaw kineticLaw = (KineticLaw) reaction.getKineticLaw();
           
             for (int k=0; k<kineticLaw.getListOfParametersSize(); k++)
             {
                 Parameter parameter = (Parameter)  kineticLaw.getListOfParameters().get(k);
                 listOfLocalParameters.add(parameter);
             }
             /** ASHISH: If reaction has been added, then all modifier-species can be made local. */
             for (int k=0; k<reaction.getListOfModifiersSize(); k++)
             {
                  ModifierSpeciesReference modifierRef = (ModifierSpeciesReference) reaction.getListOfModifiers().get(k);
		  String name = modifierRef.getSpecies();
         	  for (int l=0; l<listOfGlobalParameters.size(); l++)
         	  {
             	    Parameter globalParameter = (Parameter) listOfGlobalParameters.get(l);
		    if (name.equals(globalParameter.getId()))
		    {
			listOfLocalParameters.add(globalParameter); 
			listOfGlobalParameters.remove(l);
			break;
		    }
		  }
             } /** ASHISH: END OF LOOP */
         }
	 compartment.setLocalParameters(listOfLocalParameters);
        
         /* Fetch all the Global Parameters with Parameter-Ids different from the Parameters-Ids of Local Parameters. */   
         /* This is to ensure that Local Parameter declaration takes precedence over Global Parameter Declaration. So */
         /* if a Parameter with the same Id is declared in both Global and Local levels in SBML then the INPUT section*/
         /* in SAL MODULE will not contain the declaration correspoding to that Parameter.                            */
         
         boolean   isUniqueGlobalParameter;
         ArrayList listOfUniqueGlobalParameters = new ArrayList();

	/** The following loop is obsolete, can be removed? */
         for (int j=0; j<listOfGlobalParameters.size(); j++)
         {
             Parameter globalParameter = (Parameter) listOfGlobalParameters.get(j);
             isUniqueGlobalParameter   = true;

             for (int k=0; k<listOfLocalParameters.size(); k++)
             {
                 Parameter localParameter = (Parameter) listOfLocalParameters.get(k);
                 
                 if ( localParameter.getId().equals( globalParameter.getId() ) )
                 {
                    isUniqueGlobalParameter = false;
                    break;
                 }
             }
             
             if (isUniqueGlobalParameter)
             {
                listOfUniqueGlobalParameters.add(globalParameter);
             }    
         }
    
        /** Get all the rules (Assignment/Algerbraic/Rate) which define the constraints on the Local Species into
            temporary list Arrays.
        */
        ArrayList listOfLocalAssignmentRules = new ArrayList();
        ArrayList listOfLocalAlgebraicRules  = new ArrayList();
        ArrayList listOfLocalRateRules       = new ArrayList();
        boolean   isLocalRule                = false;
        boolean   isLocalSpecies             = false;
                  ruleVariable               = "";

        for (int j=0; j<listOfAssignmentRules.size(); j++)
        {
            Rule assignmentRule = (Rule) listOfAssignmentRules.get(j);
            isLocalRule         = false;
          
            for (int k=0; k<assignmentRule.getListOfRuleVariablesSize(); k++)
            {
                isLocalSpecies  = false;
                ruleVariable    = (String) assignmentRule.getListOfRuleVariables().get(k); 
                for (int l=0; l<listOfLocalSpecies.size(); l++)
                {
                    Species localSpecies = (Species) listOfLocalSpecies.get(l);
                    if ( ruleVariable.equals( localSpecies.getId() ) )
                    {
                       isLocalSpecies = true;
                       break;
                    }
                }
                if (isLocalSpecies)
                {
                   isLocalRule = true;
                   break;
                }
            }
            
            if (isLocalRule)
            {
               listOfLocalAssignmentRules.add(assignmentRule);
            }
        } /* Finished getting AssignmentRules coressponding to the current Module into LocalAssignmentRules ArrayList */
	 compartment.setLocalAssignmentRules(listOfLocalAssignmentRules);
      
        for (int j=0; j<listOfAlgebraicRules.size(); j++)
        {
            Rule algebraicRule  = (Rule) listOfAlgebraicRules.get(j);
            isLocalRule         = false;
          
            for (int k=0; k<algebraicRule.getListOfRuleVariablesSize(); k++)
            {
                isLocalSpecies  = false;
                ruleVariable    = (String) algebraicRule.getListOfRuleVariables().get(k); 
                for (int l=0; l<listOfLocalSpecies.size(); l++)
                {
                    Species localSpecies = (Species) listOfLocalSpecies.get(l);
                    if ( ruleVariable.equals( localSpecies.getId() ) )
                    {
                       isLocalSpecies = true;
                       break;
                    }
                }
                if (isLocalSpecies)
                {
                   isLocalRule = true;
                   break;
                }
            }
            
            if (isLocalRule)
            {
               listOfLocalAlgebraicRules.add(algebraicRule);
            }
        } /* Finished getting AlgebraicRules coressponding to the current Module into LocalAlgebraicRules ArrayList */
	 compartment.setLocalAlgebraicRules(listOfLocalAlgebraicRules);

        for (int j=0; j<listOfRateRules.size(); j++)
        {
            Rule rateRule = (Rule) listOfRateRules.get(j);
            isLocalRule  = false;
          
            for (int k=0; k<rateRule.getListOfRuleVariablesSize(); k++)
            {
                isLocalSpecies = false;
                ruleVariable   = (String) rateRule.getListOfRuleVariables().get(k); 
                for (int l=0; l<listOfLocalSpecies.size(); l++)
                {
                    Species localSpecies = (Species) listOfLocalSpecies.get(l);
                    if ( ruleVariable.equals( localSpecies.getId() ) )
                    {
                       isLocalSpecies = true;
                       break;
                    }
                }
                if (isLocalSpecies)
                {
                   isLocalRule = true;
                   break;
                }
            }
            
            if (isLocalRule)
            {
               listOfLocalRateRules.add(rateRule);
            }
        } /* Finished getting RateRules coressponding to the current Module into LocalRateRules ArrayList */
	 compartment.setLocalRateRules(listOfLocalRateRules);

        ArrayList listOfLocalEvents = new ArrayList();
        ArrayList listOfLocalPEvents = new ArrayList();
        boolean   isLocalEvent = false;
        boolean   isLocalPEvent = false;	//ASHISH: inside-out
        String    eventVariable = "";

        /** Get all the events which belong to the Compartment/Module */
        for (int j=0; j<listOfEvents.size(); j++)
        {
            Event localEvent = (Event) listOfEvents.get(j);
           
	    isLocalEvent = false;
	    isLocalPEvent = false;		//ASHISH: inside-out
            for (int k=0; k<localEvent.getListOfEventVariablesSize(); k++)
            {
                isLocalSpecies = false;
                eventVariable  = (String) localEvent.getListOfEventVariables().get(k);
                for (int l=0; l<listOfLocalSpecies.size(); l++)
                {
                    Species localSpecies = (Species) listOfLocalSpecies.get(l);
                    if ( eventVariable.equals( localSpecies.getId() ) )
                    {
                       isLocalSpecies = true;
                       break;
                    }
                }
		// ASHISH: Adding parameter check too...HERE HERE
                for (int l=0; !(isLocalSpecies) && (l<listOfLocalParameters.size()); l++)
                {
                    Parameter localParameter = (Parameter) listOfLocalParameters.get(l);
                    if ( eventVariable.equals( localParameter.getId() ) )
                    {
                       isLocalPEvent = true;
			compartment.markParameter(localParameter);	//THIS is not a const param, it changes.
                       break;
                    }
                }
                if (isLocalSpecies)
                {
                   isLocalEvent = true;
                   break;
                }
	        if (isLocalPEvent && localEvent.getListOfEventVariablesSize() == 1)
                   break;
            }
            
            if (isLocalEvent)
            {
               listOfLocalEvents.add(localEvent);
            }
            if (isLocalPEvent)
            {
               listOfLocalPEvents.add(localEvent);
            }
        }        
	compartment.setLocalEvents(listOfLocalEvents);
	compartment.setLocalPEvents(listOfLocalPEvents);
         
	/** ASH: Figure out INPUT variables from localPEvents and localEvents */
	for (int j=0; j<listOfLocalEvents.size(); j++) {
	    Event event = (Event) listOfLocalEvents.get(j);
	    ArrayList trigVars = (ArrayList) event.getListOfTriggerVariables();
	    for (int k = 0; k < event.getListOfTriggerVariablesSize(); k++) {
		compartment.addInputSpecies((String)trigVars.get(k));
	    }
	}

	for (int j=0; j<listOfLocalPEvents.size(); j++) {
	    Event event = (Event) listOfLocalPEvents.get(j);
	    ArrayList trigVars = (ArrayList) event.getListOfTriggerVariables();
	    for (int k = 0; k < event.getListOfTriggerVariablesSize(); k++) {
		compartment.addInputSpecies((String)trigVars.get(k));
	    }
	}

        /* Clear all the List arrays used to store data coressponding to a local Module. */
         
        listOfLocalParameters.clear();
        listOfLocalSpecies.clear();              
        listOfLocalReactions.clear();           
        //listOfSpeciesDot.clear();           
        listOfLocalAssignmentRules.clear();           
        listOfLocalAlgebraicRules.clear();              
        listOfLocalRateRules.clear();                   
        listOfConstantGlobalParameters.clear();              
        listOfNonConstantGlobalParameters.clear();            
        listOfUniqueGlobalParameters.clear();
        listOfLocalEvents.clear();            
          
     } /* Compartment Loop ends here */

     for (int i=0; i<listOfCompartments.size(); i++)                            // Begin of writing MODULES
     {
         Compartment compartment = (Compartment)listOfCompartments.get(i);
	 compartment.toSalFile(out, listOfGlobalParameters);
     }
         
     boolean first = true;
     if (listOfCompartments.size() > 2) {				// ASH: 1-dummy-cell;
	out.print("cell : MODULE = ");
	for (int i=0; i<listOfCompartments.size(); i++)			// cell : MODULE
	{
         Compartment compartment = (Compartment)listOfCompartments.get(i);
	 String moduleCall = compartment.getCall();
	 if ((moduleCall != null) && (!moduleCall.equals("")) && first)
	    out.print(moduleCall);
	 else if ((moduleCall != null) && (!moduleCall.equals("")))
	    out.print(" [] " + moduleCall);
	 first = false;
	}
	out.println(";\n");
     }

     if (listOfGlobalParameters.size() != 0)
     {
        out.println("ENVIRONMENT : MODULE =");
        out.println("\tBEGIN");                                                
        out.print("\tOUTPUT "); 

        for (int j=0; j<listOfGlobalParameters.size(); j++)
        {
            Parameter parameter = (Parameter)  listOfGlobalParameters.get(j);
            out.print( parameter.getId() );
            if (j < listOfGlobalParameters.size()-1)
            {
               out.print(", ");
            }
            else
            {
               out.println(" : REAL");
            }
        }
     }

     /** Get all the rules (Assignment/Algerbraic/Rate) which define the constraints on the Global Parameters into
         temporary list Arrays.
     */
     if (listOfGlobalParameters.size() != 0)                        
     {
        for (int j=0; j<listOfAssignmentRules.size(); j++)
        {
            Rule assignmentRule = (Rule) listOfAssignmentRules.get(j);
            isGlobalRule        = true;
          
            for (int k=0; k<assignmentRule.getListOfRuleVariablesSize(); k++)
            {
                isGlobalParameter = false;
                ruleVariable    = (String) assignmentRule.getListOfRuleVariables().get(k); 
                for (int l=0; l<listOfGlobalParameters.size(); l++)
                {
                    Parameter parameter = (Parameter) listOfGlobalParameters.get(l);
                    if ( ruleVariable.equals( parameter.getId() ) )
                    {
                       isGlobalParameter = true;
                    }
                }
                if (!isGlobalParameter)
                {
                   isGlobalRule = false;
                   break;
                }
            }
            
            if (isGlobalRule)
            {
               listOfGlobalAssignmentRules.add(assignmentRule);
            }
        }
      
        for (int j=0; j<listOfAlgebraicRules.size(); j++)
        {
            Rule algebraicRule  = (Rule) listOfAlgebraicRules.get(j);
            isGlobalRule        = true;
          
            for (int k=0; k<algebraicRule.getListOfRuleVariablesSize(); k++)
            {
                isGlobalParameter = false;
                ruleVariable    = (String) algebraicRule.getListOfRuleVariables().get(k); 
                for (int l=0; l<listOfGlobalParameters.size(); l++)
                {
                    Parameter parameter = (Parameter) listOfGlobalParameters.get(l);
                    if ( ruleVariable.equals( parameter.getId() ) )
                    {
                       isGlobalParameter = true;
                    }
                }
                if (!isGlobalParameter)
                {
                   isGlobalRule = false;
                   break;
                }
            }
            
            if (isGlobalRule)
            {
               listOfGlobalAlgebraicRules.add(algebraicRule);
            }
        }

        for (int j=0; j<listOfRateRules.size(); j++)
        {
            Rule rateRule = (Rule) listOfRateRules.get(j);
            isGlobalRule  = true;
          
            for (int k=0; k<rateRule.getListOfRuleVariablesSize(); k++)
            {
                isGlobalParameter = false;
                ruleVariable      = (String) rateRule.getListOfRuleVariables().get(k); 
                for (int l=0; l<listOfGlobalParameters.size(); l++)
                {
                    Parameter parameter = (Parameter) listOfGlobalParameters.get(l);
                    if ( ruleVariable.equals( parameter.getId() ) )
                    {
                       isGlobalParameter = true;
                    }
                }
                if (!isGlobalParameter)
                {
                   isGlobalRule = false;
                   break;
                }
            }
            
            if (isGlobalRule)
            {
               listOfGlobalRateRules.add(rateRule);
            }
        }

        if (listOfGlobalRateRules.size() != 0)
        {
           out.print("\tLOCAL ");
           for (int j=0; j<listOfGlobalRateRules.size(); j++)
           {
               Rule globalRateRule = (Rule) listOfGlobalRateRules.get(j);

               out.print( globalRateRule.getVariable() + "dot");

               if ( j < listOfGlobalRateRules.size()-1 )
               {
                  out.print(" ,");
               }
               else
               {
                  out.println(" : REAL");
               }
           }
        }
        
        /* Get the list of Global Parameters which will remain constant. */
        /* List of Constant Global Parameters are those for which there  */
        /* is no Global Rate Rules.                                      */

        boolean   isConstantGlobalParameter = true;

        for (int j=0; j<listOfGlobalParameters.size(); j++)
        {
            Parameter parameter = (Parameter) listOfGlobalParameters.get(j);
       
            for (int k=0; k<listOfGlobalRateRules.size(); k++)
            {
                Rule globalRateRule = (Rule) listOfGlobalRateRules.get(k);

                if ( globalRateRule.getVariable().equals(parameter.getId()) )
                {
                   isConstantGlobalParameter = false;
                   listOfNonConstantGlobalParameters.add(parameter);
                   break;      
                }
            }
           
            if (isConstantGlobalParameter) 
            {
               listOfConstantGlobalParameters.add(parameter);
            }
        }

        if ( listOfConstantGlobalParameters.size() > 0 )
        {
           out.println("\tINVARIANT");                                                
           out.print("\t       ");

           for (int j=0; j<listOfConstantGlobalParameters.size(); j++)
           {
               Parameter constantParameter = (Parameter) listOfConstantGlobalParameters.get(j);
               out.print( constantParameter.getId() + " = " + constantParameter.getValue() );

               if ( j < listOfConstantGlobalParameters.size()-1 || listOfGlobalAssignmentRules.size() > 0 ||
                    listOfGlobalAlgebraicRules.size() > 0 )
               {
                  out.print(" AND ");
               }
               else
               {
                  out.println();
               }
           }
 
        }
         
     } /* End of Global INVARIANT Section */

     if (listOfNonConstantGlobalParameters.size() != 0)
     {
        out.println("\tINITIALIZATION");
        out.print("\t       ");
        for (int j=0; j<listOfNonConstantGlobalParameters.size(); j++)
        {
            Parameter nonConstantGlobalParameter = (Parameter) listOfNonConstantGlobalParameters.get(j);
            out.print( nonConstantGlobalParameter.getId() + " = " + nonConstantGlobalParameter.getValue() );
            if (j < listOfNonConstantGlobalParameters.size()-1)
            {
               out.print("; ");
            }
            else
            {
               out.println("");
            }
        }
     } /* End of Global INITIALIZATION Section */

     if (listOfGlobalParameters.size() != 0)
     {
        out.println("\tTRANSITION");
        out.print("\t       ");
        for (int j=0; j<listOfGlobalRateRules.size(); j++)
        {
            Rule globalRateRule = (Rule) listOfGlobalRateRules.get(j);

            out.print( globalRateRule.getVariable() + "dot' = " + globalRateRule.getExpression() );

            if ( j < listOfGlobalRateRules.size()-1 )
            {
               out.println(";");
            }
            else
            {
               out.println();
            }
        }
        out.println("\tEND");
     } /* End of Global TRANSITION Section */

     out.println("END");
     System.out.println("... Finished Writing as Well");
  }

  /**
   * Make new species
   */
  public Species xmlhandlespecies(Attributes atts)
  {
       String speciesBoundaryCondition  = "false"; 
       String speciesConstant           = "false";
       String speciesInitialQuantity    = "";
       String name = atts.getValue("name");
      
       if (verbose) System.out.println("species: begin");
       if (atts.getValue("constant") != null)
       {
          speciesConstant = atts.getValue("constant");
       }
       
       if (atts.getValue("boundaryCondition") != null)
       {
          speciesBoundaryCondition = atts.getValue("boundaryCondition");
       }

       if (atts.getValue("initialAmount") != null)                 // Currently I'am treating initialAmount and Initial Conc. in
       {                                                           // the same manner as there is no unit definition structure.
          speciesInitialQuantity = atts.getValue("initialAmount"); // This blocks needs to be changed if unit definition structure
          speciesInitialQuantity = parseExpDecimal(speciesInitialQuantity); //added to SAL in future versions.
       }
       else if (atts.getValue("initialConcentration") != null)
       {
          speciesInitialQuantity = atts.getValue("initialConcentration");
          speciesInitialQuantity = parseExpDecimal(speciesInitialQuantity);
       }

       return( new Species(atts.getValue("id"), name, speciesInitialQuantity, atts.getValue("compartment"),
                                      speciesBoundaryCondition, speciesConstant ));
  }

  /**
   * This method catches <em>start of element<em> event.
   * @throws SAXException
   */
  public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
  {
    
    if (verbose) System.out.println("NEW QNAME = " + qName);
    if (equals(qName, "model"))
    {
       if (verbose) System.out.println(qName + ": begin");
       modelId = atts.getValue("id");
       if ((modelId == null) || modelId.equals("null"))
	  modelId = atts.getValue("name");
       if (modelId == null) 
	  modelId = new String("Model");
       modelId = modelId.replace(' ', '_');
    }
    
    if (equals(qName, "functionDefinition"))
    {
       if (verbose) System.out.println(qName + ": begin");
       functionId           = atts.getValue("id");
       isFunctionDefinition = true;
    }

    if (isFunctionDefinition)
    {
       if (isMath || !isMath)
       {
          if (equals(qName, "bvar"))
          {
             isFunctionVariable = true;
          }
       }
    }

    if (equals(qName, "compartment"))
    { 
       if (verbose) System.out.println(qName + ": begin");
       listOfCompartments.add(new Compartment(atts.getValue("id"),atts.getValue("spatialDimensions"),atts.getValue("name")));
    }
    
    if (equals(qName, "species"))
    {
       listOfSpecies.add( xmlhandlespecies(atts) );
    }

    if (equals(qName, "assignmentRule"))
    {
       if (verbose) System.out.println(qName + ": begin");
       ruleVariable     = atts.getValue("variable");
       listOfRuleVariables.add(ruleVariable);
       isAssignmentRule = true;
    }

    if (equals(qName, "algebraicRule"))
    {
       if (verbose) System.out.println(qName + ": begin");
       ruleVariable    = "0";
       isAlgebraicRule = true;
    }

    if (equals(qName, "rateRule"))
    {
       if (verbose) System.out.println(qName + ": begin");
       ruleVariable = atts.getValue("variable");
       listOfRuleVariables.add(ruleVariable);
       isRateRule   = true;
    }

    if ( (isAssignmentRule)     || (isAlgebraicRule) || (isRateRule) ||
         (isFunctionDefinition) || (isTrigger)       || (isEventAssignment) )
    {
       if (isMath || !isMath)
       {
          if (equals(qName, "apply"))
          {
             formula = formula + " (";
          }
          else if( equals(qName, "ci") || equals(qName, "cn") || equals(qName, "csymbol") )
          {
             mathIdentifier.setLength(0);
             isMathIdentifier = true;
          } 
       }
    } /* This block reads the MathML in each structure and constructs a PreFixExpression out of it */
   
    if (equals(qName, "reaction"))
    {
       if (verbose) System.out.println(qName + ": begin");
       reactionName       = atts.getValue("id");
       reactionReversible = atts.getValue("reversible");
    }
    
    if (equals(qName, "listOfReactants"))
    {
       if (verbose) System.out.println(qName + ": begin");
       isListOfReactants = true;
    }
    
    if (equals(qName, "listOfProducts"))
    {
       if (verbose) System.out.println(qName + ": begin");
       isListOfProducts = true;
    }

    if (equals(qName, "listOfModifiers"))
    {
       if (verbose) System.out.println(qName + ": begin");
       isListOfModifiers = true;
    }

    if (equals(qName, "speciesReference"))
    {
       if (verbose) System.out.println(qName + ": begin");
       speciesValue       = atts.getValue("species");
       stoichiometryValue = "1"; 
       isSpeciesReference = true;
    }
    
    if (equals(qName, "modifierSpeciesReference"))
    {
       if(isListOfModifiers)
       {
          listOfModifiers.add(new ModifierSpeciesReference(atts.getValue("species")));
       }
    }

    if (equals(qName, "math"))
    {
       if (verbose) System.out.println(qName + ": begin");
       isMath = true;
    }

    if (isSpeciesReference)
    {
       if (isMath)
       {
          if (equals(qName, "cn"))
          {
             mathIdentifier.setLength(0);
             isMathIdentifier = true;
          }
       }
    }
    
    if (equals(qName, "kineticLaw"))
    {
       if (verbose) System.out.println(qName + ": begin");
       isKineticLaw = true;
    }

    if (isKineticLaw)
    { 
       if (isMath)
       {
          if (equals(qName, "apply"))
          {
             formula = formula + " (";
          }
          else if( equals(qName, "ci") || equals(qName, "cn") || equals(qName, "csymbol") )
          {
             mathIdentifier.setLength(0);
             isMathIdentifier = true;
          }
       }
       if (equals(qName, "parameter"))
       {
          String parameterValue = parseExpDecimal(atts.getValue("value"));
          listOfParameters.add(new Parameter(atts.getValue("id"),parameterValue));
       }
    }
    
    if ( (!isKineticLaw) && equals(qName, "parameter") )
    {
       String parameterValue = parseExpDecimal(atts.getValue("value"));
       listOfGlobalParameters.add(new Parameter(atts.getValue("id"),parameterValue));
    }

    if (equals(qName, "trigger"))
    {
       if (verbose) System.out.println(qName + ": begin");
       isTrigger = true;
    }

    if (equals(qName, "eventAssignment") || equals(qName, "assignment"))		//ASHISH: adding option
    {
       if (verbose) System.out.println(qName + ": begin");
       isEventAssignment = true;
       eventVariable     = atts.getValue("variable");
       listOfEventVariables.add(eventVariable);
    }

  }
 
  /**
   * This method catches <em>end of element<em> event.
   * @throws SAXException
   */
  public void endElement(String uri, String localName, String qName) throws SAXException
  {
   
    if (verbose) System.out.println("END QNAME = " + qName);
    if (equals(qName, "assignmentRule")) 
    {
       formula          = Prefix_To_Infix_Converter(formula);
       listOfAssignmentRules.add( new Rule(ruleVariable, formula, listOfRuleVariables) );
       formula          = "";
       isAssignmentRule = false;
       listOfRuleVariables.clear();
    }  /* Converts the AssignmentRule from the Prefix form to Infix form and adds it to ArrayList */

    if (equals(qName, "algebraicRule"))
    {
       formula         = Prefix_To_Infix_Converter(formula);
       listOfAlgebraicRules.add( new Rule(ruleVariable, formula, listOfRuleVariables) );
       formula         = "";
       isAlgebraicRule = false;
       listOfRuleVariables.clear();
    }  /* Converts the AlgebraicRule from the Prefix form to Infix form and adds it to ArrayList */

    if (equals(qName, "rateRule"))
    {
       formula    = Prefix_To_Infix_Converter(formula);
       listOfRateRules.add( new Rule(ruleVariable, formula, listOfRuleVariables) );
       formula    = "";
       isRateRule = false;
       listOfRuleVariables.clear();
    }  /* Converts the rateRule from the Prefix form to Infix form and adds it to ArrayList */
   
    if (isSpeciesReference)
    {
       if (isMath)
       {
          if ( equals(qName, "ci") || equals(qName, "cn") )
          {
             isMathIdentifier   = false;
             stoichiometryValue = "";
             stoichiometryValue = stoichiometryValue + mathIdentifier;
             stoichiometryValue = stoichiometryValue.trim();
             stoichiometryValue = parseExpDecimal(stoichiometryValue);
          }
       }
       
    }
       
    if (equals(qName, "speciesReference"))
    {
       isSpeciesReference = false; 
 
       if (isListOfReactants)
       {
          listOfReactants.add(new SpeciesReference(speciesValue,stoichiometryValue));
       }
       
       if (isListOfProducts)
       {
          listOfProducts.add(new SpeciesReference(speciesValue,stoichiometryValue));
       }

    }

    if (equals(qName, "listOfReactants"))
    {
       isListOfReactants = false;
    }

    if (equals(qName, "listOfProducts"))
    {
       isListOfProducts = false;
    }
   
    if (equals(qName, "listOfModifiers"))
    {
       isListOfModifiers = false;
    }

    if (equals(qName, "kineticLaw"))
    {
       if ((formula != null) && !(formula.equals("")))
          formula      = Prefix_To_Infix_Converter(formula);
       else
          formula      = "";
       isKineticLaw = false;
    }

    // if ( (isTrigger) || (isEventAssignment) )
    if ( (isEventAssignment) )						// ASHISH: Removing isTrigger.
    {
       if (equals(qName, "ci"))                                            // This block keeps track of all the variables
       {                                                                  // involved in Event Assignments;        
          String Variable = "";
          Variable = Variable + mathIdentifier;
          Variable = Variable.trim();
          listOfEventVariables.add(Variable);
       }
    }

    if ( (isTrigger) )							// ASHISH: Removing isTrigger.
    {
       if (equals(qName, "ci"))                                            // This block keeps track of all the variables
       {                                                                  // involved in Event Assignments;        
          String Variable = "";
          Variable = Variable + mathIdentifier;
          Variable = Variable.trim();
          listOfTriggerVariables.add(Variable);
       }
    }
   
    if ( (isKineticLaw) || (isAssignmentRule) || (isAlgebraicRule)   ||
         (isRateRule)   || (isTrigger)        || (isEventAssignment) ||
         (isFunctionDefinition && !isFunctionVariable) )
    {
       if (isMath || !isMath)
       {
          
          if (equals(qName, "ci"))
          {                                                                 
             if ( (isRateRule) || (isAssignmentRule) || (isAlgebraicRule) )         // This block keeps track of all the variables
             {                                                                      // involved in different types of rules;
                String Variable = "";
                Variable = Variable + mathIdentifier;
                Variable = Variable.trim();
                listOfRuleVariables.add(Variable);
             }
          }

          if ( equals(qName, "ci") || equals(qName, "cn") || equals(qName, "csymbol") ) // This block parsers the mathML expressions
          {
             isMathIdentifier = false;
             String Identifier = "";
             Identifier = Identifier + mathIdentifier;
             Identifier = Identifier.trim();
             
             if (equals(qName, "cn"))
             {
                Identifier = parseExpDecimalPreFix(Identifier);                     // Convert the numbers to Power10-PreFix
                                                                                    // format. Which will be converted back to
                                                                                    // InFixFormat later.
             }            
              
             formula          = formula + " " + Identifier;
          }  
          if (equals(qName, "apply"))
          {
             formula = formula + " )";
          }
          if (equals(qName, "times"))
          {
             formula = formula + " *";
          }
          if (equals(qName, "divide"))
          {
             formula = formula + " /";
          }
          if (equals(qName, "plus"))
          {
             formula = formula + " +";
          }
          if (equals(qName, "minus"))
          {
             formula = formula + " -";
          }
          if (equals(qName, "power"))
          {
             formula = formula + " ^";
          }
          if (equals(qName, "eq"))
          {
             formula = formula + " =";
          }
          if (equals(qName, "neq"))
          {
             formula = formula + " !=";
          }
          if (equals(qName, "gt"))
          {
             formula = formula + " >";
          }
          if (equals(qName, "geq"))
          {
             formula = formula + " >=";
          }
          if (equals(qName, "lt"))
          {
             formula = formula + " <";
          }
          if (equals(qName, "leq"))
          {
             formula = formula + " <=";
          }
          if (equals(qName, "and"))
          {
             formula = formula + " AND ";
          }
       }
    }
    
    if (isFunctionDefinition && isFunctionVariable)
    {  
       if (isMath || !isMath)
       {
          if ( equals(qName, "ci") )
          {
             isMathIdentifier = false;
             functionVariable = "";
             functionVariable = functionVariable + mathIdentifier;
             functionVariable = functionVariable.trim();
             listOfFunctionVariables.add( functionVariable );
          }  
       } 
    }
    
    if (equals(qName, "functionDefinition"))
    {
       isFunctionDefinition = false;
       formula              = Prefix_To_Infix_Converter(formula);
       listOfFunctionDefinitions.add( new FunctionDefinition(functionId,listOfFunctionVariables,formula) );
       listOfFunctionVariables.clear();
       formula              = ""; 
    }
       
    if (equals(qName, "bvar"))
    {
       isFunctionVariable   = false;
    }

    if (equals(qName, "trigger"))
    {  
       if (verbose) System.out.println("exiting trigger 1");
       if (verbose) System.out.println("formula = " + formula);
       triggerExpression    = "";
       if (formula != null && !(formula.equals(""))) {
	  System.out.println("formula in prefix is " + formula);
          triggerExpression    = Prefix_To_Infix_Converter(formula);
       } else
          triggerExpression    = "FALSE";
       formula              = "";
       isTrigger            = false;
       if (verbose) System.out.println("exiting trigger 2");
    }

    if (equals(qName, "eventAssignment") || equals(qName, "assignment"))	//ASHISH: Adding new option.
    {
       formula           = Prefix_To_Infix_Converter(formula);
       eventAssignment   = eventVariable + "' = " + formula;			//ASHISH: NEXT operator
       listOfEventAssignments.add(eventAssignment);
       formula           = "";
       isEventAssignment = false;
    }

    if (equals(qName, "event"))
    {
       if (verbose) System.out.println("out of event 1");
       if (!triggerExpression.equals("FALSE"))					//ASHISH: condition added.
       	  listOfEvents.add( new Event(triggerExpression, listOfEventAssignments, listOfEventVariables, listOfTriggerVariables) );
       if (listOfEventAssignments != null)
          listOfEventAssignments.clear();
       if (listOfEventVariables != null)
          listOfEventVariables.clear();
       if (listOfTriggerVariables != null)
          listOfTriggerVariables.clear();
       if (verbose) System.out.println("out of event 2");
    }

    if (equals(qName, "math"))
    {
       isMath = false;
    }

    if (equals(qName, "reaction"))
    {
        Reaction r   = new Reaction(reactionName,reactionReversible);
        r.cloneListOfReactants(listOfReactants);
        r.cloneListOfProducts(listOfProducts);
        r.cloneListOfModifiers(listOfModifiers);
        KineticLaw k = new KineticLaw(formula);
        k.cloneListOfParameters(listOfParameters);
        r.cloneKineticLaw(k);

        listOfReactants.clear();
        listOfProducts.clear();
        listOfModifiers.clear();
        listOfParameters.clear();
        formula = "";

        listOfReactions.add(r);
    }
       
  }
 
  public void characters(char[] chars, int start, int len) throws SAXException
  {
    if (isMathIdentifier)
    {
       mathIdentifier.append(chars, start, len);
    }
  }

  /** Converts a Prefix Expression to Infix Expression
   *  Note: The tokens in the input string should be separated by blank space.
   *  @return InfixExpression
   */
  public String Prefix_To_Infix_Converter (String prefixExpression)
  {
    Stack expressionStack = new Stack();
    StringTokenizer t     = new StringTokenizer(prefixExpression);
    String operator;
 
    do
    {
      Object element = new String(t.nextToken());
 
      if ( element.equals(")") )
      {
         Stack tempTermStack = new Stack();
         element             = expressionStack.pop();
 
         while ( !element.equals("(") )
         {
           tempTermStack.push(element);
           element = (String) expressionStack.pop();
         }
 
         Object term = new String("(");
         operator    = (String)tempTermStack.pop();

         if ( operator.equals("*")  || operator.equals("/")  || operator.equals("+")  ||
              operator.equals("-")  || operator.equals("^")  || operator.equals(">")  ||
              operator.equals("<")  || operator.equals("<=") || operator.equals(">=") ||
              operator.equals("!=") || operator.equals("=")  || operator.equals("AND"))
         {  
            while (!tempTermStack.empty())
            {
              term = term + (String)tempTermStack.pop();
             
              if (!tempTermStack.empty())
              {
                 term = term + " " + operator + " ";	//ASHISH: Adding " "
              }
           
            }
         }   
         else
         { 
            term = operator + term;             // In this case operator refers to function name. Also
                                                // operator is already initialized to left paranthesis
                                                // so we can concatenate function name and "(" and assign
                                                // it back to term.
            while (!tempTermStack.empty())
            {
              term = term + (String)tempTermStack.pop();
              
              if (!tempTermStack.empty())
              {
                 term = term + ",";
              }
            }
         }

         term = term + ")";
         expressionStack.push(term);
 
       }
       else
       {
         expressionStack.push(element);
       }

     }while(t.hasMoreTokens());
 
     String expression = (String)expressionStack.pop();
     return expression;
  }

  /** This method converts a String (numbers) containing decimals and 'e' to an equivalent expression in powers of 10. 
   *  For example 2.345e-2 will be converted to 2345 * (1/1000) * (1/100)
   *  Note: This method inturn makes a call to parseDecimal method to convert Strings containing Decimal numbers to
   *        an equivalent expression in powers of 10.
   *  @return powerTenExpression
   */
  public String parseExpDecimal(String oldString)
  {
     int posOfExp     = oldString.indexOf('e');
     if (posOfExp >= 0)
     {
        if (oldString.charAt(posOfExp+1) == '-')
        {
           String beginStr  = oldString.substring(0,posOfExp);
           beginStr         = decimalParser(beginStr);
           String endStr    = oldString.substring(posOfExp+2,oldString.length());
           String tenPower  = "1";
           for (int i=0; i < Integer.parseInt(endStr); i++)
           {
               tenPower = tenPower + "0";
           }
           String newString = beginStr + " * (1/" + tenPower + ")";
           return newString;
        }
        else
        {
           String beginStr  = oldString.substring(0,posOfExp);
           beginStr         = decimalParser(beginStr);
           String endStr    = oldString.substring(posOfExp+1,oldString.length());
           String tenPower  = "1";
           for (int i=0; i < Integer.parseInt(endStr); i++)
           {
               tenPower = tenPower + "0";
           }
           String newString = beginStr + " * (" + tenPower + ")";
           return newString;
          
        }
     }
     else
     {
        String newString = decimalParser(oldString);
        return newString;
     }  
  }  
 
  /** This method converts Strings containing Decimal numbers to an equivalent expression in powers of 10.
   *  For Example 2.345 will be converted to 2345 * (1/1000)      
   */
  public String decimalParser(String oldString)
  {
     int posOfDecimal =  oldString.indexOf('.');

     if (posOfDecimal >= 0)
     {
        int strLength    = oldString.length();
        String beginStr  = oldString.substring(0,posOfDecimal);
        String endStr    = oldString.substring(posOfDecimal+1,strLength);
        String tenPower  = "1";
        for (int i=0; i < endStr.length(); i++)
        {
            tenPower = tenPower + "0";
        }
        String newString = beginStr + endStr + " * (1/" + tenPower + ")";
        return newString;
     }
     else
     {
        return oldString;
     }
  }

  /** This method converts a String (numbers) containing decimals and 'e' to an equivalent expression in powers of 10. 
   *  For example 2.345e-2 will be converted to ( * 2345 ( / 1 1000 ) ( / 1 100 ) ) 
   *  Note: This method inturn makes a call to parseDecimal method to convert Strings containing Decimal numbers to
   *        an equivalent expression in powers of 10.
   *  @return powerTenExpression
   */
  public static String parseExpDecimalPreFix(String oldString)
  {
     int posOfExp     = oldString.indexOf('e');
     if (posOfExp >= 0)
     {
        if (oldString.charAt(posOfExp+1) == '-')
        {
           String beginStr  = oldString.substring(0,posOfExp);
           beginStr         = decimalParserPreFix(beginStr);
           String endStr    = oldString.substring(posOfExp+2,oldString.length());
           String tenPower  = "1";
           for (int i=0; i < Integer.parseInt(endStr); i++)
           {
               tenPower = tenPower + "0";
           }
           String newString = "( * " + beginStr + " ( / 1 " + tenPower + " ) " + ")";
           return newString;
        }
        else
        {
           String beginStr  = oldString.substring(0,posOfExp);
           beginStr         = decimalParserPreFix(beginStr);
           String endStr    = oldString.substring(posOfExp+1,oldString.length());
           String tenPower  = "1";
           for (int i=0; i < Integer.parseInt(endStr); i++)
           {
               tenPower = tenPower + "0";
           }
           String newString = "( * " + beginStr + " ( " + tenPower + " ) " + ")";
           return newString;
          
        }
     }
     else
     {
        String newString = decimalParserPreFix(oldString);
        if (newString.equals(oldString))
        {
           return oldString;
        }
        else
        {
           return "( * " + newString + " )"; 
        }
     }  
  }  
 
  /** This method converts Strings containing Decimal numbers to an equivalent expression in powers of 10.
   *  For Example 2.345 will be converted to 2345 ( / 1 1000 )       
   */
  public static String decimalParserPreFix(String oldString)
  {
     int posOfDecimal =  oldString.indexOf('.');

     if (posOfDecimal >= 0)
     {
        int strLength    = oldString.length();
        String beginStr  = oldString.substring(0,posOfDecimal);
        String endStr    = oldString.substring(posOfDecimal+1,strLength);
        String tenPower  = "1";
        for (int i=0; i < endStr.length(); i++)
        {
            tenPower = tenPower + "0";
        }
        String newString = beginStr + endStr + " ( / 1 " + tenPower + " )";
        return newString;
     }
     else
     {
        return oldString;
     }
  }

  /**
   * This method is used to check whether the document is read properly in the corresponding fields or not.
   * This method "can" be removed once the program is completed.
   */
  public void printCompartments()
  {
    System.out.println("Model Id: " + modelId);


    for (int j=0; j<listOfFunctionDefinitions.size(); j++)
    {
        FunctionDefinition f = (FunctionDefinition)  listOfFunctionDefinitions.get(j);
        System.out.println("Function Definition Id:" + f.getFunctionId() + " FunctionDefinitionExpression: "
                                                     + f.getFunctionExpression());
        for (int i=0; i<f.getListOfFunctionVariablesSize(); i++)
        {
            System.out.println( "variable" + i + ":" + f.getListOfFunctionVariables().get(i) );
        }
    }

    for (int j=0; j<listOfGlobalParameters.size(); j++)
    {
        Parameter p = (Parameter)  listOfGlobalParameters.get(j);
        System.out.println("Parameter-Id: " + p.getId() + "Value: " + p.getValue());
    }

    for(int i=0; i<listOfCompartments.size(); i++)
    {
       Compartment c = (Compartment)listOfCompartments.get(i);
       System.out.println("Compartment Id: " + c.getId()
                          + ", spatialDimensions: " + c.getSpatialDimensions());
    }
    
    for(int i=0; i<listOfSpecies.size(); i++)
    {
       Species s = (Species)listOfSpecies.get(i);
       System.out.println("Species Id: " + s.getId() + ", InitialAmount: " + s.getInitialAmount() + ", Compartment: "
                          + s.getCompartment() + ", boundaryCondition: " + s.getBoundaryCondition() + ", Constant: "
                          + s.getConstant());
    }

    for (int j=0; j<listOfAssignmentRules.size(); j++)
    {
        Rule l = (Rule) listOfAssignmentRules.get(j);
        System.out.println("Assignment-RHS-Expression: " + l.getExpression() + "Variable: " + l.getVariable());
   
        System.out.println("list of Assignment Rule Variables: ");
        for (int i=0; i<l.getListOfRuleVariablesSize(); i++)
        {
            System.out.print(l.getListOfRuleVariables().get(i) + " ");
        }
        System.out.println(); 
    }
      
    for (int j=0; j<listOfAlgebraicRules.size(); j++)
    {
        Rule l = (Rule) listOfAlgebraicRules.get(j);
        System.out.println("Algebraic-RHS-Expression: " + l.getExpression() + "Variable: " + l.getVariable());

        System.out.println("list of Algebraic Rule Variables: ");
        for (int i=0; i<l.getListOfRuleVariablesSize(); i++)
        {
            System.out.print(l.getListOfRuleVariables().get(i) + " ");
        }
        System.out.println(); 
             
    }

    for (int j=0; j<listOfRateRules.size(); j++)
    {
        Rule l = (Rule) listOfRateRules.get(j);
        System.out.println("Rate-RHS-Expression: " + l.getExpression() + "Variable: " + l.getVariable());

        System.out.println("list of Rate Rule Variables: ");
        for (int i=0; i<l.getListOfRuleVariablesSize(); i++)
        {
            System.out.print(l.getListOfRuleVariables().get(i) + " ");
        }
        System.out.println(); 
             
    }

    for(int i=0; i<listOfReactions.size();i++)
    {

       Reaction r = (Reaction)listOfReactions.get(i);
       System.out.println("Reaction Id: " + r.getId() + ", Reversible: " +  r.getReversible());
       
       for (int j=0; j<r.getListOfReactantsSize();j++)
       {
           SpeciesReference s = (SpeciesReference) r.getListOfReactants().get(j);
           System.out.println("Reactant Id: " +  s.getSpecies() + ", Stoichiometry: " + s.getStoichiometry());
       }

       for (int j=0; j<r.getListOfProductsSize();j++)
       {
           SpeciesReference s = (SpeciesReference) r.getListOfProducts().get(j);
           System.out.println("Product Id: " +  s.getSpecies() + ", Stoichiometry: " + s.getStoichiometry());
       }
       
       for (int j=0; j<r.getListOfModifiersSize();j++)
       {
           ModifierSpeciesReference m = (ModifierSpeciesReference) r.getListOfModifiers().get(j);
           System.out.println("Modifier Id: " +  m.getSpecies());
       }

       KineticLaw k = (KineticLaw) r.getKineticLaw();
       System.out.println("Kinetic Law-Formula: " + k.getFormula());
       
       for (int j=0; j<k.getListOfParametersSize(); j++)
       {
           Parameter p = (Parameter)  k.getListOfParameters().get(j);
           System.out.println("Parameter-Id: " + p.getId() + "Value: " + p.getValue());
       }
       
    }

    for (int j=0; j<listOfEvents.size(); j++)
    {
        Event e = (Event) listOfEvents.get(j);
        System.out.println("Trigger:" + e.getTriggerExpression()); 
        for (int i=0; i<e.getListOfEventAssignmentsSize(); i++)
        {
            System.out.println( "Event Assignment" + i + ":" + e.getListOfEventAssignments().get(i) );
        }
        for (int i=0; i<e.getListOfEventVariablesSize(); i++)
        {
            System.out.println( "Event Variable"   + i + ":" + e.getListOfEventVariables().get(i) );
        }
    }

  }

  /* Check if qName is equal to something:string. 
   * Everything before ":" is thus ignored in qName. */
  private boolean equals(String qName, String str)
  {
	int i = qName.lastIndexOf(':');
	boolean ret = false;
	if (i == -1) {
	   ret = qName.equals(str);
	} else if (!(qName.startsWith("sbml") || qName.startsWith("math"))) {
	   ret = false;
	} else if (qName.endsWith(str)) {
	   if (verbose) System.out.println("qName = " + qName);
	   if (verbose) System.out.println("Index = " + String.valueOf(i));
	   if (verbose) System.out.println("str = " + str);
	   if (verbose) System.out.println("len = " + str.length());
	   ret = qName.regionMatches(i+1, str, 0, str.length());
	   if (verbose) System.out.println("ret = " + String.valueOf(ret));
	}
	return(ret);
  }

}
