package sbmlClasses;
import java.util.*;
/**
 * A <code>Reaction</code> represents any transformation, transport or binding process, typically
 * a chemical reaction, that can transform or change the amount of one or more species. A reaction
 * is defined primarily in terms of the participating reactants and products (and their correspon-
 * ding stoichiometries), along with optional modifier species, an optional kinetic law describing
 * the rate at which the reaction takes place, and optional parameters entering into the kinetic 
 * law.
 */ 
public class Reaction 
{
  /**
   * This Constructor initializes the id and reversible parameters.
   * @param id          Indicates a unique identifier for the Reaction.
   * @param reversible  Indicates Reaction is reversible or not.
   */
  public Reaction(String id, String reversible)
  {
    this.id = id;
    this.reversible = reversible;
  }

  /**
   * This method gets the unique Identifier of the reaction.
   * @return Id       Reaction Identifier
   */
  public String getId()
  {
    return id;
  }
  
  /**
   * This method returns a String(not boolean) "true" if the reaction is reversible else it returns "false".
   * @return reversible "true" if reaction is reversible else "false".
   */
  public String getReversible()
  {
    return reversible;
  }

  /**
   * This method makes a shallow copy of all the <code>reactants</code> corresponding to the 
   * <code>reaction</code> from the temporary ListArray to the ListArray declared in reaction object.
   * @param listOfReactants List of all the reactants corresponding to a particular reaction.
   */
  public void cloneListOfReactants(ArrayList listOfReactants)
  {
     this.listOfReactants = (ArrayList) listOfReactants.clone();// Note: A Shallow Copy is enough in this case
                                                                //       since the contents of this ArrayList
                                                                //       are immutable.
  }
  
  /**
   * This method fetches the number of reactants participating in a reaction. 
   * @return int Size of the Array list in which reactants are stored.
   */
  public int getListOfReactantsSize()
  {
     return listOfReactants.size();
  }

  /**
   * This method fetches the Array list of reactants participating in a reaction. 
   * @return ArrayList List of the reactants participating in a reaction.
   */
  public ArrayList getListOfReactants()
  {
     return listOfReactants;
  }

  /**
   * This method makes a shallow copy of all the <code>products</code> corresponding to the 
   * <code>reaction</code> from the temporary ListArray to the ListArray declared in reaction object.
   * @param listOfProducts List of all the Products corresponding to a reaction.
   */
  public void cloneListOfProducts(ArrayList listOfProducts)
  {
     this.listOfProducts = (ArrayList) listOfProducts.clone();// Note: A Shallow Copy is enough in this case
                                                              //       since the contents of this ArrayList
                                                              //       are immutable.
  }
  
  /**
   * This method fetches the number of products participating in a reaction. 
   * @return int Size of the Array list in which products are stored.
   */
  public int getListOfProductsSize()
  {
     return listOfProducts.size();
  }

  /**
   * This method fetches the Array list of products participating in a reaction. 
   * @return ArrayList List of the products generated in a reaction.
   */
  public ArrayList getListOfProducts()
  {
     return listOfProducts;
  }

  /**
   * This method makes a shallow copy of all the <code>modifiers</code> corresponding to the 
   * <code>reaction</code> from the temporary ListArray to the ListArray declared in reaction object.
   * @param listOfModifiers List of all the Modifiers corresponding to a reaction.
   */
  public void cloneListOfModifiers(ArrayList listOfModifiers)
  {
     this.listOfModifiers = (ArrayList) listOfModifiers.clone();// Note: A Shallow Copy is enough in this case
                                                                //       since the contents of this ArrayList
                                                                //       are immutable.
  }
  
  /**
   * This method fetches the number of modifiers participating in a reaction. 
   * @return int Size of the Array list in which modifiers are stored.
   */
  public int getListOfModifiersSize()
  {
     return listOfModifiers.size();
  }

  /**
   * This method fetches the Array list of modifierss participating in a reaction. 
   * @return ArrayList List of the modifiers generated in a reaction.
   */
  public ArrayList getListOfModifiers()
  {
     return listOfModifiers;
  }

  /**
   * Copies the data from a temporary KineticLaw object to the KineticLaw in the reaction object.
   * @param KineticLaw contains kinetic information corresponding to a reaction.
   */ 
  public void cloneKineticLaw(KineticLaw kineticLaw)
  {
     this.kineticLaw = (KineticLaw) kineticLaw.clone();
  }

  /**
   * This method fethces the kinetic information corresponding to a reaction object.
   * @return KineticLaw contains kinetic information corresponding to a reaction.
   */
  public KineticLaw getKineticLaw()
  {
     return kineticLaw;
  }

  private String id;
  private String reversible;
  private ArrayList listOfReactants = new ArrayList();
  private ArrayList listOfProducts  = new ArrayList();
  private ArrayList listOfModifiers = new ArrayList();
  private KineticLaw kineticLaw     = new KineticLaw();
}
